/*
  VIA GPIO kernel driver code
    K-TEAM SA,
    
    2010.02.09, J. Tharin: first version
    
    
*/


/********************************************************************
* 
* Included headers
*
*/


#include <linux/init.h>  /* Needed for the macros */

#include <linux/module.h> /* Needed by all modules */

#include <linux/kernel.h> // printk


#include <asm/io.h> /* inb, outb */

#include <linux/pci.h> // pci_read_word,...

#include <asm/uaccess.h> /* copy_from/to_user */

#include <linux/fs.h> // register_chrdev




#include "via_gpio.h"


MODULE_LICENSE("GPL");

/********************************************************************
* 
* Definitions
*
*/

#undef DEBUG



// CX700/VX700 
#define VIA_GPIO_VENDOR 0x1106

#define VIA_GPIO_CLASS 0x0601

#define VIA_GPIO_DEVICE_ID 0x8324


#define VIA_GPIO_PCI_BUS 0x0

#define VIA_GPIO_PCI_DEVICE 0x11

#define VIA_GPIO_PCI_FUNCTION 0x0

//  PCI domain (host bridge)
#define VIA_GPIO_PCI_DOMAIN 0x0 


#define PMIO_ADDRESS 0x400

/********************************************************************
* 
* Functions declaration
*
*/

static int via_gpio_ioctl(
  struct inode* inode,
  struct file* file,
  unsigned int cmd,
  unsigned long arg);


/********************************************************************
* 
* Global variables
*
*/


uint16_t pmio_address;


struct pci_dev *via_gpio_pci_dev = NULL;


/* Structure that declares the common */
/* file access fcuntions */
struct file_operations via_gpio_fops = {
 ioctl:via_gpio_ioctl
};



/********************************************************************
* 
* Functions definitions
*
*/


/********************************************************************
* 
* Internal functions
*
*/


// set bit at address
static void bit_set(unsigned long addr,unsigned bit)
{
 
  uint8_t val;
  
  #ifdef DEBUG
  printk(KERN_INFO "via_gio bit_set addr: 0x%x  | bit: %d\n",addr,bit);
  #endif
  
  val = inb(pmio_address+addr);
  val |= 1<<bit; // set to 1
  outb(val,pmio_address+addr);
 
  
  
}


// clear bit at address
static void bit_clear(unsigned long addr,unsigned bit)
{
  uint8_t val;

  #ifdef DEBUG
  printk(KERN_INFO "via_gio bit_clear addr: 0x%x  | bit: %d\n",addr,bit);
  #endif
 
  val = inb(pmio_address+addr);
  val &= ~(1<<bit);
  outb(val,pmio_address+addr);

  
}


// test bit at address
static int bit_test(unsigned long addr,unsigned bit)
{
  uint8_t val; 
   
  val = inb(pmio_address+addr);
  
    
  #ifdef DEBUG
  printk(KERN_INFO "via_gio bit_test addr: 0x%x  | bit: %d | val: %d\n",addr,bit,val & (2<<bit));
  #endif
  
  if(val & (1<<bit))
    return 1;
  else
    return 0;
}


/*!
 * Initialize the GPIO module. This function must be called before any
 * other call the gpio functions.
 *
 * \return 
 * 	- 0 on success
 *	- a negative error code on failure
 */
static int gpios_init()
{
	int i;
  uint8_t val;
  struct pci_dev  *gpio_dev;


  gpio_dev=via_gpio_pci_dev;
  
  if (gpio_dev == NULL)
  {
    printk(KERN_ALERT "ERROR: CX700/VX700 PCI to ISA Bridge not found: cannot get pci dev!\n"); 
    return -1;
  }
  

  // initialise GPIO2
  /*  GPIO2
	control register bit: B0D17F0 Rx53[7]=0
	status register : PMIO Rx49[6]
	Output register: PMIO Rx4D[6]
	Input register: PMIO Rx49[6] ? */

  // control register
  pci_read_config_byte(gpio_dev, 0x53,&val);
	val &= ~0b10000000;
  pci_write_config_byte(gpio_dev, 0x53, &val);



  // initialise GPIO3
/* GPIO3
	control register bit: B0D17F0  Rx53[7]=0
	status register: PMIO Rx49[7]
	Output register: PMIO Rx4D[7]
	Input register: PMIO Rx49[7] ? */
  
   // control register
  
  pci_read_config_byte(gpio_dev, 0x53,&val);
	val &= ~0b10000000;
  pci_write_config_byte(gpio_dev, 0x53, &val);

  // initialise GPI4   
 /*GPI4 :
	control register bit: PMIO Rx25[4]=0 & PMIO Rx23[4]=0
	status register: PMIO Rx48[4] 
	Input register:  PMIO Rx48[4] ? */
	
  // control registers
   
  bit_clear(0x23,4);
  bit_clear(0x25,4);  
   

  // initialise GPI5     
  /*GPI5
	control register bit: PMIO Rx24[4]=0 & PMIO Rx22[4]=0
	status register: PMIO Rx48[5]
	Input register:  PMIO Rx48[5] ? */
	
   // control registers

  bit_clear(0x22,4);
  bit_clear(0x24,4); 

  
   

  return 0;
}


/********************************************************************
* 
* Module functions
*
*/


// init function
static int __init via_gpio_init(void)
{
  
  int result;

	printk(KERN_INFO "Loading via_gpio module\n"); 

  /* Registering device */
  result = register_chrdev(VIA_GPIO_MAJOR, VIA_GPIO_DEVICE_NAME, 
      &via_gpio_fops);
  if (result < 0) { 
    printk(KERN_ALERT "via_gpio: cannot obtain major number %d\n",VIA_GPIO_MAJOR); 
    return result; 
  } 
   

  // get pci

	via_gpio_pci_dev=pci_find_device(VIA_GPIO_VENDOR,VIA_GPIO_DEVICE_ID,NULL);
  
  
  if (via_gpio_pci_dev == NULL)
  {   
    printk(KERN_ALERT "\nERROR: CX700/VX700 PCI to ISA Bridge not found: cannot get pci dev!\n");
    return -1;
	 }
	
	
 
  #ifdef DEBUG
  pci_read_config_word(via_gpio_pci_dev, 0x0,&pmio_address) ;
  printk(KERN_INFO "Vendor ID = 0x%x\n",pmio_address);
  
  pci_read_config_word(via_gpio_pci_dev, 0x2,&pmio_address) ;
  printk(KERN_INFO "Device ID = 0x%x\n",pmio_address);
  #endif
 

  /* We need the I/O Base Address for this board.
  p 151. "CX700 Series System Programming Manual"
  
  bits 7-> 15 => mask 80 for 7 and ff for 8 to 15
	        
 */
  pci_read_config_byte(via_gpio_pci_dev, 0x88,&pmio_address);
  #ifdef DEBUG
  printk(KERN_INFO "via_gpia: 0x88 address value for PMIO  = 0x%x\n",pmio_address);
  #endif
  
  pmio_address &= 0xff80;

  #ifdef DEBUG
  printk(KERN_INFO "via_gpia: pmio base address = 0x%x\n",pmio_address);
	#endif

  // getting address from the 0x88 address of the pci board above seems not to work in kernel mode
  pmio_address=PMIO_ADDRESS; 
  
  #ifdef DEBUG
  printk(KERN_INFO "via_gpia: default pmio base address = 0x%x\n",pmio_address);
  #endif
  
  gpios_init();


	printk(KERN_INFO "via_gpio module loaded successfully\n");
  return 0;

  
}

// exit function
static void __exit via_gpio_exit(void)
{
   /* Make major number free! */
  unregister_chrdev(VIA_GPIO_MAJOR, VIA_GPIO_DEVICE_NAME);


  printk(KERN_INFO "Releasing via_gpio module\n");

}


// main function
static int via_gpio_ioctl(
  struct inode* inode,
  struct file* file,
  unsigned int cmd,
  unsigned long arg)
{
  t_gpio_port gpio_port;
  int rval =0;

 
  /* Make sure that we can access the data
  * from the user application */
  if(!access_ok(VERIFY_WRITE, (void *)arg, sizeof(t_gpio_port)))
  {
  	printk(KERN_ALERT "GPIO: cannot read argument from user application!\n");
  	return -EBUSY;
  }
  
    
  #ifdef DEBUG
  printk(KERN_INFO "via_gpio_ioctl access \n");
  #endif
  
  /* Make a local copy of the data from the
  * user application */
  copy_from_user(&gpio_port,(void*)arg, sizeof(t_gpio_port));
  
  
  switch(cmd){
    case GPIO_TEST:
      gpio_port.value=bit_test(gpio_port.address,gpio_port.bit);
      /* copy the value read back to
      * the user application */
      copy_to_user((void*)arg, &gpio_port, sizeof(t_gpio_port));
      #ifdef DEBUG1
      printk(KERN_INFO "GPIO_TEST : ad 0x%x | bit %d | val %d\n",gpio_port.address,gpio_port.bit,gpio_port.value);
      #endif
    break;
    case GPIO_SET:
      /* write the value to the port */
      bit_set(gpio_port.address,gpio_port.bit);
      #ifdef DEBUG1
      printk(KERN_INFO "GPIO_SET : ad 0x%x | bit %d\n",gpio_port.address,gpio_port.bit);
      #endif
      break;
    case GPIO_CLEAR:
    /* write configuration settings to
      * the configuration register */
      bit_clear(gpio_port.address,gpio_port.bit);
      #ifdef DEBUG1
      printk(KERN_INFO "GPIO_CLEAR : ad 0x%x | bit %d\n",gpio_port.address,gpio_port.bit);
      #endif
      break;
    default:
      /* Return an error since the command
      * is not recognized */
      printk(KERN_ALERT "GPIO: Unknown IOCTL %x", cmd);
      return -EINVAL;
  }
  
  
  return rval;
}


module_init(via_gpio_init);
module_exit(via_gpio_exit);
